import { Message, TextBasedChannel } from "discord.js";
import { Character, CHARACTERS, Effect, PotencyStatus, Target } from "./characters";
import { Player } from "./models/player";
import { Unit } from "./models/unit";

export enum BattleState {
    Ongoing,
    Team1Win,
    Team2Win
}

interface StatusEffects {
    poison: number
    regeneration: number
    burn: number
    confusion: number
    stun: number
}

interface PotencyEffects {
    resistanceChange: PotencyStatus[]
    accuracyChange: PotencyStatus[]
    speedChange: PotencyStatus[]
    damageChange: PotencyStatus[]
}

class BattleUnit {
    constructor (unit: Unit) {
        this.unit = unit;
        this.defaultHealth = unit.health;
        this.health = unit.health;
        this.speed = unit.speed;
        this.character = CHARACTERS.get(unit.character_id)!;
    }

    unit: Unit
    defaultHealth: number
    health: number
    speed: number
    character: Character
    statusEffects: StatusEffects = {
        poison: 0,
        regeneration: 0,
        burn: 0,
        confusion: 0,
        stun: 0
    }
    potencyEffects: PotencyEffects = {
        resistanceChange: [],
        accuracyChange: [],
        speedChange: [],
        damageChange: []
    }
    active = true
}

export class Battle {
    constructor (channel: TextBasedChannel, player1: Player, player2: Player, team1: [Unit, Unit, Unit], team2: [Unit, Unit, Unit]) {
        this.channel = channel;
        this.player1 = player1;
        this.player2 = player2;
        this.team1 = team1.map(unit => {return new BattleUnit(unit)}) as [BattleUnit, BattleUnit, BattleUnit];
        this.team2 = team2.map(unit => {return new BattleUnit(unit)}) as [BattleUnit, BattleUnit, BattleUnit];
        this.units = this.team1.concat(this.team2) as [BattleUnit, BattleUnit, BattleUnit, BattleUnit, BattleUnit, BattleUnit];
    }

    simulateRound (): BattleState {
        this.log = '';
        this.tickStatuses(this.units);
        this.checkAlive(this.units);
        this.useSkills(this.units);

        if (!this.team1.filter(e => {return e.active})[0]) return BattleState.Team2Win;
        else if (!this.team2.filter(e => {return e.active})[0]) return BattleState.Team1Win;
        else return BattleState.Ongoing;
    }

    tickStatuses (units: BattleUnit[]) {
        for (const unit of units) {
            if (!unit.active) continue;

            if (unit.statusEffects.poison > 0) {
                unit.statusEffects.poison--
                const change = Math.floor(unit.defaultHealth/12);
                unit.health -= change;
                this.appendLog(`${unit.character.nameShort} took ${change} poison damage!`);
            }
            this.log += '\n';
            if (unit.statusEffects.regeneration > 0) {
                unit.statusEffects.regeneration--
                const change = Math.floor(unit.defaultHealth/10)
                unit.health += change;
                this.appendLog(`${unit.character.nameShort} regenerated ${change} health!`);
            }
            this.log += '\n';
            if (unit.statusEffects.burn > 0) {
                unit.statusEffects.burn--
                unit.health -= 5;
                this.appendLog(`${unit.character.nameShort} took 5 burn damage!`);
            }
            this.log += '\n';
        }
    }

    useSkills (units: BattleUnit[]) {
        const order = [];

        for (let i = 0; i < units.length; i++) {
            const rolls = []
            let modifiedSpeed = units[i].speed
            for (const change of units[i].potencyEffects.speedChange) {
                modifiedSpeed -= change.potency;
            }
            for (let o = 0; o < modifiedSpeed; o++) {
                rolls.push(Math.random());
            }
            const initiative = Math.max(...rolls);
            order.push([i, initiative]);
        }

        order.sort((a, b) => {return b[1] - a[1]});

        console.debug(order);

        for (const initiative of order) {
            const unit = units[initiative[0]];
            if (!unit.active) continue;

            const skill = unit.character.skills[Math.floor(Math.random()*unit.character.skills.length)];
            const team = initiative[0] < 3 ? 1 : 2
            const target = team === 1 ? this.team2[Math.floor(Math.random()*this.team2.length)] : this.team1[Math.floor(Math.random()*this.team1.length)];
            const allyTarget = team === 1 ? this.team1[Math.floor(Math.random()*this.team1.length)] : this.team2[Math.floor(Math.random()*this.team2.length)];

            if (skill.accuracy && Math.random()*100 > skill.accuracy) {
                this.appendLog(`${unit.character.nameShort} used ${skill.name}, but missed!`);
            }

            this.appendLog(`${unit.character.nameShort} used ${skill.name}!`);
            for (const effect of skill.effects) {
                switch (effect.target){
                    case Target.Self:
                    this.runEffect(effect, unit, [unit]);
                    break;
                    case Target.Team:
                    this.runEffect(effect, unit, team === 1 ? this.team1 : this.team2);
                    break;
                    case Target.OneTeammate:
                    this.runEffect(effect, unit, [allyTarget]);
                    break;
                    case Target.OneOpponent:
                    this.runEffect(effect, unit, [target]);
                    break;
                    case Target.AllOpponents:
                    this.runEffect(effect, unit, team === 1 ? this.team2 : this.team1);
                    break;
                    case Target.TrueRandomOpponent:
                    this.runEffect(effect, unit, [team === 1 ? this.team2[Math.floor(Math.random()*this.team2.length)] : this.team1[Math.floor(Math.random()*this.team1.length)]]);
                    break;
                }
            }
            this.checkAlive(this.units);
        }
    }

    runEffect (effect: Effect, self: BattleUnit, targets: BattleUnit[]) {
        if (effect.damage) {
            for (const target of targets) {
                let damage = effect.damage;
                for (const statChange of self.potencyEffects.damageChange) {
                    Math.round(damage *= 1+(statChange.potency/100));
                }
                for (const statChange of target.potencyEffects.resistanceChange) {
                    Math.round(damage *= 1-(statChange.potency/100));
                }

                this.appendLog(`${target.character.nameShort} took ${damage} damage!`);
                target.health -= damage;
            }
        }
    }

    checkAlive (units: BattleUnit[]) {
        for (const unit of units) {
            if (!unit.active) continue;
            console.debug(unit.health);
            if (unit.health > 0) continue;
            unit.active = false;
            this.appendLog(`${unit.character.nameShort} has been defeated!`);
        }
    }

    appendLog (str: string) {
        if (this.log) this.log += '\n';
        this.log += str;
        console.debug(str);
        return this.log;
    }

    message?: Message
    log = ""

    channel: TextBasedChannel
    player1: Player
    player2: Player
    team1: [BattleUnit, BattleUnit, BattleUnit]
    team2: [BattleUnit, BattleUnit, BattleUnit]
    units: [BattleUnit, BattleUnit, BattleUnit, BattleUnit, BattleUnit, BattleUnit]
}